/* /////////////////////////////////////////////////////////////////////////////
 * File:		perf.d
 *
 * Purpose:		Win32 performance types
 *
 * Created		15th March 2003
 * Updated:		3rd April 2003
 *
 * Copyright:	Synesis Software Pty Ltd, (c) 2002. All rights reserved.
 *
 * Home:		http://www.synesis.com.au/software
 *
 * ////////////////////////////////////////////////////////////////////////// */



/* ////////////////////////////////////////////////////////////////////////// */

module synsoft.win32.perf;

/* ////////////////////////////////////////////////////////////////////////// */

import windows;

/* ////////////////////////////////////////////////////////////////////////// */

extern (Windows)
{
	BOOL QueryPerformanceCounter(long *);
	BOOL QueryPerformanceFrequency(long *);
	DWORD GetTickCount();
	BOOL GetProcessTimes(HANDLE hThread, FILETIME *lpCreationTime, FILETIME *lpExitTime, FILETIME *lpKernelTime, FILETIME *lpUserTime);
	BOOL GetThreadTimes(HANDLE hThread, FILETIME *lpCreationTime, FILETIME *lpExitTime, FILETIME *lpKernelTime, FILETIME *lpUserTime);
	HANDLE GetCurrentThread();
	HANDLE GetCurrentProcess();
}

/* ////////////////////////////////////////////////////////////////////////// */

/// \brief Performance counter scope class generating template
///
/// This template declares the class Scope which manages a performance counter 
/// instance, scoping its measurement interval by virtue of its supporting the
/// RAII protocol.
template PerformanceCounterScope(T)
{
	/// \brief Performance counter scoping class
	///
	/// This auto class manages a performance counter instance, scoping its
	/// measurement interval by virtue of its supporting the RAII protocol
	auto class Scope
	{
	private:
		alias Scope     class_type;
	public:
		/// The parameterising performance counter class
		alias T         counter_type;

	public:
		/// Constructs an instance of the PerformanceCounterScope using a reference 
		/// to the parameterising performance counter class, whose measurement is
		/// to be scoped. The measurement interval is commenced with a call to 
		/// <b>start()</b> on the managed counter instance.
		this(counter_type counter)
		{
			m_counter = counter;

			m_counter.start();
		}
		/// The measurement interval is completed with a call to <b>stop()</b> on the 
		/// managed counter instance.
		~this()
		;

		/// Calls <b>stop()</b> on the managed counter instance, so that intermediate
		/// timings can be taken.
		void stop()
		;

		/// Returns a reference to the managed counter instance.
		counter_type get_counter()
		;

	// Members
	private:
		T   m_counter;

	// Not to be implemented
	private:
		this(class_type rhs);
//		class_type operator =(class_type rhs);
	}
}


/* ////////////////////////////////////////////////////////////////////////// */

/// \brief A performance counter that uses the high performance hardware counter on the host machine
///
/// This class provides high-resolution performance monitoring using the host machine's high performance
/// hardware counter. This class does not provide meaningful timing information on operating systems
/// that do not provide a high performance hardware counter.
class HighPerformanceCounter
{
private:
	alias	long	epoch_type;
public:
	/// \brief The interval type
	///
	/// The type of the interval measurement, a 64-bit signed integer
	alias	long	interval_type;

	/// \brief The scope type
	///
	/// The type with which instances of the counter class can be subject to the RAII protocol in order
	/// to scope particular measurement intervals
	alias instance PerformanceCounterScope(HighPerformanceCounter).Scope	scope_type;

// Constructors
private:
	;

// Operations
public:
	/// \brief Starts measurement
	///
	/// Begins the measurement period
    void start()
	;
	/// \brief Ends measurement
	///
	/// Ends the measurement period
    void stop()
	;

// Attributes
public:
	/// \brief The elapsed count in the measurement period
	///
	/// This represents the extent, in machine-specific increments, of the measurement period
    interval_type get_period_count()
	;
	/// \brief The number of whole seconds in the measurement period
	///
	/// This represents the extent, in whole seconds, of the measurement period
    interval_type get_seconds()
	;
	/// \brief The number of whole milliseconds in the measurement period
	///
	/// This represents the extent, in whole milliseconds, of the measurement period
    interval_type get_milliseconds()
	;
	/// \brief The number of whole microseconds in the measurement period
	///
	/// This represents the extent, in whole microseconds, of the measurement period
    interval_type get_microseconds()
	;

// Members
private:
    epoch_type				m_start;    // start of measurement period
    epoch_type				m_end;      // End of measurement period
	static interval_type	sm_freq;	// Frequency
}

/* ////////////////////////////////////////////////////////////////////////// */

/// \brief A low-cost, low-resolution performance counter
///
/// This class provides low-resolution, but low-latency, performance monitoring
/// and is guaranteed to be meaningful on all operating systems.
class TickCounter
{
private:
	alias	long	epoch_type;
public:
	/// \brief The interval type
	///
	/// The type of the interval measurement, a 64-bit signed integer
	alias	long	interval_type;

	/// \brief The scope type
	///
	/// The type with which instances of the counter class can be subject to the RAII protocol in order
	/// to scope particular measurement intervals
	alias instance PerformanceCounterScope(TickCounter).Scope	scope_type;

// Constructors
public:

// Operations
public:
	/// \brief Starts measurement
	///
	/// Begins the measurement period
    void start()
	;
	/// \brief Ends measurement
	///
	/// Ends the measurement period
    void stop()
	;

// Attributes
public:
	/// \brief The elapsed count in the measurement period
	///
	/// This represents the extent, in machine-specific increments, of the measurement period
    interval_type get_period_count()
	;
	/// \brief The number of whole seconds in the measurement period
	///
	/// This represents the extent, in whole seconds, of the measurement period
    interval_type get_seconds()
	;
	/// \brief The number of whole milliseconds in the measurement period
	///
	/// This represents the extent, in whole milliseconds, of the measurement period
    interval_type get_milliseconds()
	;
	/// \brief The number of whole microseconds in the measurement period
	///
	/// This represents the extent, in whole microseconds, of the measurement period
    interval_type get_microseconds()
	;

// Members
private:
    uint	m_start;    // start of measurement period
    uint	m_end;      // End of measurement period
}

/* ////////////////////////////////////////////////////////////////////////// */

/// \brief A performance counter that provides thread-specific performance timings
///
/// This class uses the operating system's performance monitoring facilities to provide timing 
/// information pertaining to the calling thread only, irrespective of the activities of other
/// threads on the system. This class does not provide meaningful timing information on operating
/// systems that do not provide thread-specific monitoring.
class ThreadTimesCounter
{
private:
	alias	long	epoch_type;
public:
	/// \brief The interval type
	///
	/// The type of the interval measurement, a 64-bit signed integer
	alias	long	interval_type;

	/// \brief The scope type
	///
	/// The type with which instances of the counter class can be subject to the RAII protocol in order
	/// to scope particular measurement intervals
	alias instance PerformanceCounterScope(ThreadTimesCounter).Scope	scope_type;

// Constructors
public:
	/// \brief Constructor
	///
	/// Creates an instance of the class, and caches the thread token so that measurements will
	/// be taken with respect to the thread in which the class was created.
	this()
	{
	    m_thread = GetCurrentThread();
	}

// Operations
public:
	/// \brief Starts measurement
	///
	/// Begins the measurement period
    void start()
	;
	/// \brief Ends measurement
	///
	/// Ends the measurement period
    void stop()
	;

// Attributes
public:

    // Kernel

	/// \brief The elapsed count in the measurement period for kernel mode activity
	///
	/// This represents the extent, in machine-specific increments, of the measurement period for kernel mode activity
    interval_type get_kernel_period_count()
	;
	/// \brief The number of whole seconds in the measurement period for kernel mode activity
	///
	/// This represents the extent, in whole seconds, of the measurement period for kernel mode activity
    interval_type get_kernel_seconds()
	;
	/// \brief The number of whole milliseconds in the measurement period for kernel mode activity
	///
	/// This represents the extent, in whole milliseconds, of the measurement period for kernel mode activity
    interval_type get_kernel_milliseconds()
	;
	/// \brief The number of whole microseconds in the measurement period for kernel mode activity
	///
	/// This represents the extent, in whole microseconds, of the measurement period for kernel mode activity
    interval_type get_kernel_microseconds()
	;

    // User

	/// \brief The elapsed count in the measurement period for user mode activity
	///
	/// This represents the extent, in machine-specific increments, of the measurement period for user mode activity
    interval_type get_user_period_count()
	;
	/// \brief The number of whole seconds in the measurement period for user mode activity
	///
	/// This represents the extent, in whole seconds, of the measurement period for user mode activity
    interval_type get_user_seconds()
	;
	/// \brief The number of whole milliseconds in the measurement period for user mode activity
	///
	/// This represents the extent, in whole milliseconds, of the measurement period for user mode activity
    interval_type get_user_milliseconds()
	;
	/// \brief The number of whole microseconds in the measurement period for user mode activity
	///
	/// This represents the extent, in whole microseconds, of the measurement period for user mode activity
    interval_type get_user_microseconds()
	;

    // Total

	/// \brief The elapsed count in the measurement period
	///
	/// This represents the extent, in machine-specific increments, of the measurement period
    interval_type get_period_count()
	;
	/// \brief The number of whole seconds in the measurement period
	///
	/// This represents the extent, in whole seconds, of the measurement period
    interval_type get_seconds()
	;
	/// \brief The number of whole milliseconds in the measurement period
	///
	/// This represents the extent, in whole milliseconds, of the measurement period
    interval_type get_milliseconds()
	;
	/// \brief The number of whole microseconds in the measurement period
	///
	/// This represents the extent, in whole microseconds, of the measurement period
    interval_type get_microseconds()
	;

// Members
private:
    epoch_type  m_kernelStart;
    epoch_type  m_kernelEnd;
    epoch_type  m_userStart;
    epoch_type  m_userEnd;
    HANDLE      m_thread;
}

/* ////////////////////////////////////////////////////////////////////////// */

/// \brief A performance counter that provides process-specific performance timings
///
/// This class uses the operating system's performance monitoring facilities to provide timing 
/// information pertaining to the calling process only, irrespective of the activities of other
/// processes on the system. This class does not provide meaningful timing information on operating
/// systems that do not provide process-specific monitoring.
class ProcessTimesCounter
{
private:
	alias	long	epoch_type;
public:
	/// \brief The interval type
	///
	/// The type of the interval measurement, a 64-bit signed integer
	alias	long	interval_type;

	/// \brief The scope type
	///
	/// The type with which instances of the counter class can be subject to the RAII protocol in order
	/// to scope particular measurement intervals
	alias instance PerformanceCounterScope(ProcessTimesCounter).Scope	scope_type;

// Constructors
private:
	/// \brief Class constructor
	///
	/// Detects availability of the high performance hardware counter, and if
	/// not available adjusts 
	;

// Operations
public:
	/// \brief Starts measurement
	///
	/// Begins the measurement period
    void start()
	;
	/// \brief Ends measurement
	///
	/// Ends the measurement period
    void stop()
	;

// Attributes
public:

    // Kernel

	/// \brief The elapsed count in the measurement period for kernel mode activity
	///
	/// This represents the extent, in machine-specific increments, of the measurement period for kernel mode activity
    interval_type get_kernel_period_count()
	;
	/// \brief The number of whole seconds in the measurement period for kernel mode activity
	///
	/// This represents the extent, in whole seconds, of the measurement period for kernel mode activity
    interval_type get_kernel_seconds()
	;
	/// \brief The number of whole milliseconds in the measurement period for kernel mode activity
	///
	/// This represents the extent, in whole milliseconds, of the measurement period for kernel mode activity
    interval_type get_kernel_milliseconds()
	;
	/// \brief The number of whole microseconds in the measurement period for kernel mode activity
	///
	/// This represents the extent, in whole microseconds, of the measurement period for kernel mode activity
    interval_type get_kernel_microseconds()
	;

    // User

	/// \brief The elapsed count in the measurement period for user mode activity
	///
	/// This represents the extent, in machine-specific increments, of the measurement period for user mode activity
    interval_type get_user_period_count()
	;
	/// \brief The number of whole seconds in the measurement period for user mode activity
	///
	/// This represents the extent, in whole seconds, of the measurement period for user mode activity
    interval_type get_user_seconds()
	;
	/// \brief The number of whole milliseconds in the measurement period for user mode activity
	///
	/// This represents the extent, in whole milliseconds, of the measurement period for user mode activity
    interval_type get_user_milliseconds()
	;
	/// \brief The number of whole microseconds in the measurement period for user mode activity
	///
	/// This represents the extent, in whole microseconds, of the measurement period for user mode activity
    interval_type get_user_microseconds()
	;

    // Total

	/// \brief The elapsed count in the measurement period
	///
	/// This represents the extent, in machine-specific increments, of the measurement period
    interval_type get_period_count()
	;
	/// \brief The number of whole seconds in the measurement period
	///
	/// This represents the extent, in whole seconds, of the measurement period
    interval_type get_seconds()
	;
	/// \brief The number of whole milliseconds in the measurement period
	///
	/// This represents the extent, in whole milliseconds, of the measurement period
    interval_type get_milliseconds()
	;
	/// \brief The number of whole microseconds in the measurement period
	///
	/// This represents the extent, in whole microseconds, of the measurement period
    interval_type get_microseconds()
	;

// Members
private:
    epoch_type		m_kernelStart;
    epoch_type		m_kernelEnd;
    epoch_type		m_userStart;
    epoch_type		m_userEnd;
    static HANDLE   sm_process;
}

/* ////////////////////////////////////////////////////////////////////////// */

/// \brief A performance counter that uses the most accurate measurement APIs available on the host machine
///
/// This class attempts to use the high performance hardware counter as its measurement resource, but failing 
/// that it defaults to less accurate resources in order to guarantee that meaningful measurements are always
/// available to application code
class PerformanceCounter
{
private:
	alias	long	epoch_type;
public:
	/// \brief The interval type
	///
	/// The type of the interval measurement, a 64-bit signed integer
	alias	long	interval_type;

	/// \brief The scope type
	///
	/// The type with which instances of the counter class can be subject to the RAII protocol in order
	/// to scope particular measurement intervals
	alias instance PerformanceCounterScope(PerformanceCounter).Scope	scope_type;

// Constructors
private:
	/// \brief Class constructor
	///
	/// Detects availability of the high performance hardware counter, and if
	/// not available adjusts 
	;

// Operations
public:
	/// \brief Starts measurement
	///
	/// Begins the measurement period
    void start()
	;
	/// \brief Ends measurement
	///
	/// Ends the measurement period
    void stop()
	;

// Attributes
public:
	/// \brief The elapsed count in the measurement period
	///
	/// This represents the extent, in machine-specific increments, of the measurement period
    interval_type get_period_count()
	;
	/// \brief The number of whole seconds in the measurement period
	///
	/// This represents the extent, in whole seconds, of the measurement period
    interval_type get_seconds()
	;
	/// \brief The number of whole milliseconds in the measurement period
	///
	/// This represents the extent, in whole milliseconds, of the measurement period
    interval_type get_milliseconds()
	;
	/// \brief The number of whole microseconds in the measurement period
	///
	/// This represents the extent, in whole microseconds, of the measurement period
    interval_type get_microseconds()
	;


// Implementation
private:
	alias void function(out epoch_type interval)	measure_func;

	static void _qpc(out epoch_type interval)
	;

	static void _qtc(out epoch_type interval)
	;

// Members
private:
    epoch_type				m_start;    // start of measurement period
    epoch_type				m_end;      // End of measurement period
	static interval_type	sm_freq;	// Frequency
	static measure_func		sm_fn;		// Measurement function
}

/* ////////////////////////////////////////////////////////////////////////// */
